/*
 * fullscr.cpp - determine whether there is a fullscreen application active
 * Copyright (c) 2005, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#undef WINVER
#define WINVER 0x0500

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <ddraw.h>

#include "iat.h"

#define NHOOKS 2

static BOOL WINAPI mySetForegroundWindow(HWND hWnd) {
  // don't do anything
  return TRUE;
}

static VOID WINAPI myMouseEvent(DWORD dwFlags, DWORD dx, DWORD dy,
 DWORD dwData, DWORD dwExtraInfo) {
  // don't do anything
}

static iat_hooks sfw_hook[NHOOKS] = {
  { "SetForegroundWindow", -1, mySetForegroundWindow },
  { "mouse_event", -1, myMouseEvent }
};

typedef HRESULT (WINAPI *DIRECTDRAWCREATE)(GUID FAR *lpGUID,
   LPDIRECTDRAW FAR *lplpDD, IUnknown FAR *pUnkOuter);

static BOOL isDirectXActive(HWND mWnd) {
  DIRECTDRAWCREATE pDirectDrawCreate;
  LPDIRECTDRAW lpDD;
  HINSTANCE hlibDDraw;
  HRESULT res;

  // load ddraw.dll
  hlibDDraw = LoadLibrary("ddraw.dll");
  if (hlibDDraw == NULL) return FALSE;

  // hook the SetForegroundWindow function, to prevent activation of
  // our window when we call the SetCooperativeLevel function below
  iat_hook(hlibDDraw, "user32.dll", sfw_hook, NHOOKS);

  // get a pointer to DirectDrawCreate
  pDirectDrawCreate =
   (DIRECTDRAWCREATE)GetProcAddress(hlibDDraw, "DirectDrawCreate");
  if (pDirectDrawCreate == NULL) {
    FreeLibrary(hlibDDraw);

    return FALSE;
  }

  // create a DirectDraw instance
  if ((res = pDirectDrawCreate(NULL, &lpDD, NULL)) != DD_OK) {
    FreeLibrary(hlibDDraw);

    return FALSE;
  }

  // see if we can enter exclusive mode
  res = lpDD->SetCooperativeLevel(mWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN |
   DDSCL_NOWINDOWCHANGES);

  // if so, quickly change back!
  if (res == DD_OK) lpDD->SetCooperativeLevel(mWnd, DDSCL_NORMAL);

  // get rid of the DirectDraw instance
  lpDD->Release();

  // and ddraw.dll (no need to call iat_unhook first)
  FreeLibrary(hlibDDraw);

  // did setting exclusive mode fail because of another application?
  return (res == DDERR_EXCLUSIVEMODEALREADYSET);
}

extern "C" int __export __stdcall directx(HWND mWnd, HWND aWnd, char *data,
 char *parms, BOOL show, BOOL nopause) {
  wsprintf(data, "%d", isDirectXActive(mWnd));

  return 3;
}

static BOOL isCandidate(HWND hwndFore) {
  if (!IsWindowVisible(hwndFore) || IsIconic(hwndFore)) return FALSE;

  if (GetWindowLong(hwndFore, GWL_STYLE) & WS_BORDER) return FALSE;

  return TRUE;
}

static BOOL coversDesktop(HWND hwndFore) {
  RECT rectFore, rectDesk, rectSect;
  HWND hwndDesk;

  hwndDesk = GetDesktopWindow();
  if (hwndFore == hwndDesk) return FALSE;

  GetWindowRect(hwndFore, &rectFore);
  GetWindowRect(hwndDesk, &rectDesk);
  IntersectRect(&rectSect, &rectFore, &rectDesk);

  return EqualRect(&rectSect, &rectDesk);
}

static BOOL isExplorer(HWND hwndFore) {
  DWORD pidFore, pidProg;
  HWND hwndProg;

  hwndProg = FindWindow("PROGMAN", "Program Manager");

  GetWindowThreadProcessId(hwndFore, &pidFore);
  GetWindowThreadProcessId(hwndProg, &pidProg);

  return pidFore == pidProg;
}

static BOOL isTopmost(HWND hwndFore) {
  if (GetWindowLong(hwndFore, GWL_EXSTYLE) & WS_EX_TOPMOST) return TRUE;

  return FALSE;
}

extern "C" int __export __stdcall fullscr(HWND mWnd, HWND aWnd, char *data,
 char *parms, BOOL show, BOOL nopause) {
  HWND hwndFore;
  int res;

  hwndFore = GetForegroundWindow();

  res = isCandidate(hwndFore);
  if (res) res = coversDesktop(hwndFore);
  if (res) res = !isExplorer(hwndFore);
  if (res && isTopmost(hwndFore)) res++;

  wsprintf(data, "%d", res);

  return 3;
}

static BOOL isScreensaverActive(void) {
  BOOL bActive;

  if (SystemParametersInfo(SPI_GETSCREENSAVERRUNNING, 0, &bActive, 0))
   return bActive;

  return FALSE;
}

extern "C" int __export __stdcall scrsave(HWND mWnd, HWND aWnd, char *data,
 char *parms, BOOL show, BOOL nopause) {
  wsprintf(data, "%d", isScreensaverActive());

  return 3;
}
