/*
 * iat.c - in-process Import Address Table function hooking
 * Copyright (c) 2005-2006, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "iat.h"

static BOOL write_data(void *ptr, void *src, DWORD size) {
  MEMORY_BASIC_INFORMATION mbi;
  DWORD prot;

  VirtualQuery(ptr, &mbi, sizeof(mbi));

  if (!VirtualProtect(mbi.BaseAddress, mbi.RegionSize, PAGE_READWRITE, &prot))
   return FALSE;

  memcpy(ptr, src, size);

  VirtualProtect(mbi.BaseAddress, mbi.RegionSize, prot, &prot);

  return TRUE;
}

static BOOL set_hook(iat_hooks *hooks, int nhooks, char *name, int ordinal,
 DWORD *ipt) {
  int i;

  // find a match for this IAT entry, in the list of functions to hook
  for (i = 0; i < nhooks; i++) {
    // check for a name match
    if (name != NULL && hooks[i].name != NULL &&
     !lstrcmpi(name, hooks[i].name)) break;

    // check for an ordinal match
    if (hooks[i].ordinal >= 0 && ordinal == hooks[i].ordinal) break;
  }

  if (i == nhooks) return FALSE;

  // hook the function
  hooks[i].next = (void *)*ipt;

  if (!write_data(ipt, &hooks[i].proc, sizeof(void *))) return FALSE;

  hooks[i].ptr = ipt;

  return TRUE;
}

int iat_hook(HINSTANCE source, char *import, iat_hooks *hooks, int nhooks) {
  PIMAGE_DOS_HEADER dos;
  PIMAGE_NT_HEADERS nt;
  PIMAGE_IMPORT_DESCRIPTOR desc;
  DWORD *iat, *ipt;
  char *base, *name;
  int ordinal, hits = 0;

  // get from the program's base address to its import table
  base = (char *)source;
  dos = (PIMAGE_DOS_HEADER)base;

  if (dos->e_magic != IMAGE_DOS_SIGNATURE) return 0;

  nt = (PIMAGE_NT_HEADERS)(base + dos->e_lfanew);

  if (nt->Signature != IMAGE_NT_SIGNATURE) return 0;

  desc = (PIMAGE_IMPORT_DESCRIPTOR)(base + nt->OptionalHeader.
   DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);

  // find the given import module in the list
  for ( ; desc->Name != NULL; desc++)
   if (!lstrcmpi(base + desc->Name, import)) break;

  if (desc->Name == NULL) return 0;

  // walk through the IAT entries
  iat = (DWORD *)(base + desc->OriginalFirstThunk);
  ipt = (DWORD *)(base + desc->FirstThunk);

  for ( ; *iat && *ipt; iat++, ipt++) {
    name = NULL;
    ordinal = -1;

    // get the IAT entry's function name or ordinal
    if (IMAGE_SNAP_BY_ORDINAL(*iat)) ordinal = IMAGE_ORDINAL(*iat);
    else name = ((PIMAGE_IMPORT_BY_NAME)(base + *iat))->Name;

    // check if either of these match the given entries; if so, hook them
    if (set_hook(hooks, nhooks, name, ordinal, ipt)) hits++;
  }

  // return the number of functions we successfully hooked
  return hits;
}

int iat_unhook(iat_hooks *hooks, int nhooks) {
  int i, hits = 0;

  // for each successfully hooked function, try to unhook it
  for (i = 0; i < nhooks; i++) {
    if (hooks[i].ptr != NULL) {
      // make sure we're not unhooking in the middle of a chain
      if (*(DWORD *)hooks[i].ptr != (DWORD)hooks[i].proc) continue;

      // write back the next function in the function chain
      if (!write_data(hooks[i].ptr, &hooks[i].next, sizeof(void *)))
       continue;

      hooks[i].ptr = NULL;
      hits++;
    }
  }

  // return the number of functions we successfully unhooked
  return hits;
}
